@@ -10,15 +10,12 @@ import (
 	"fmt"
 	"io"
 	"io/ioutil"
-	"math"
 	"net/http"
 	"net/url"
 	"path"
 	"sort"
-	"strconv"
 	"strings"
 	"sync"
-	"time"
 
 	"github.com/go-kit/kit/log"
 	"github.com/go-kit/kit/log/level"
@@ -26,56 +23,48 @@ import (
 	"github.com/golang/snappy"
 	"github.com/opentracing/opentracing-go"
 	"github.com/pkg/errors"
-	"github.com/prometheus/common/version"
+	"github.com/prometheus/client_golang/prometheus"
+	"github.com/prometheus/client_golang/prometheus/promauto"
 	"github.com/prometheus/prometheus/pkg/labels"
-	"github.com/prometheus/prometheus/pkg/timestamp"
 	"github.com/prometheus/prometheus/storage/remote"
 	"github.com/prometheus/prometheus/tsdb/chunkenc"
+	"github.com/thanos-io/thanos/pkg/store/labelpb"
+	"google.golang.org/grpc/codes"
+	"google.golang.org/grpc/status"
+
 	"github.com/thanos-io/thanos/pkg/component"
-	"github.com/thanos-io/thanos/pkg/exthttp"
+	thanoshttp "github.com/thanos-io/thanos/pkg/http"
+	"github.com/thanos-io/thanos/pkg/promclient"
 	"github.com/thanos-io/thanos/pkg/runutil"
 	"github.com/thanos-io/thanos/pkg/store/storepb"
 	"github.com/thanos-io/thanos/pkg/store/storepb/prompb"
 	"github.com/thanos-io/thanos/pkg/tracing"
-	"google.golang.org/grpc/codes"
-	"google.golang.org/grpc/status"
 )
 
-var statusToCode = map[int]codes.Code{
-	http.StatusBadRequest:          codes.InvalidArgument,
-	http.StatusNotFound:            codes.NotFound,
-	http.StatusUnprocessableEntity: codes.Internal,
-	http.StatusServiceUnavailable:  codes.Unavailable,
-	http.StatusInternalServerError: codes.Internal,
-}
-
-var userAgent = fmt.Sprintf("Thanos/%s", version.Version)
-
 // PrometheusStore implements the store node API on top of the Prometheus remote read API.
 type PrometheusStore struct {
 	logger         log.Logger
 	base           *url.URL
-	client         *http.Client
+	client         *promclient.Client
 	buffers        sync.Pool
 	component      component.StoreAPI
 	externalLabels func() labels.Labels
 	timestamps     func() (mint int64, maxt int64)
 
 	remoteReadAcceptableResponses []prompb.ReadRequest_ResponseType
+
+	framesRead prometheus.Histogram
 }
 
-const (
-	initialBufSize = 32 * 1024 // 32KB seems like a good minimum starting size.
-
-	SUCCESS = "success"
-)
+const initialBufSize = 32 * 1024 // 32KB seems like a good minimum starting size for sync pool size.
 
 // NewPrometheusStore returns a new PrometheusStore that uses the given HTTP client
 // to talk to Prometheus.
 // It attaches the provided external labels to all results.
 func NewPrometheusStore(
 	logger log.Logger,
-	client *http.Client,
+	reg prometheus.Registerer,
+	client *promclient.Client,
 	baseURL *url.URL,
 	component component.StoreAPI,
 	externalLabels func() labels.Labels,
@@ -84,11 +73,6 @@ func NewPrometheusStore(
 	if logger == nil {
 		logger = log.NewNopLogger()
 	}
-	if client == nil {
-		client = &http.Client{
-			Transport: tracing.HTTPTripperware(logger, exthttp.NewTransport()),
-		}
-	}
 	p := &PrometheusStore{
 		logger:                        logger,
 		base:                          baseURL,
@@ -101,6 +85,13 @@ func NewPrometheusStore(
 			b := make([]byte, 0, initialBufSize)
 			return &b
 		}},
+		framesRead: promauto.With(reg).NewHistogram(
+			prometheus.HistogramOpts{
+				Name:    "prometheus_store_received_frames",
+				Help:    "Number of frames received per streamed response.",
+				Buckets: prometheus.ExponentialBuckets(10, 10, 5),
+			},
+		),
 	}
 	return p, nil
 }
@@ -108,28 +99,23 @@ func NewPrometheusStore(
 // Info returns store information about the Prometheus instance.
 // NOTE(bwplotka): MaxTime & MinTime are not accurate nor adjusted dynamically.
 // This is fine for now, but might be needed in future.
-func (p *PrometheusStore) Info(ctx context.Context, r *storepb.InfoRequest) (*storepb.InfoResponse, error) {
+func (p *PrometheusStore) Info(_ context.Context, _ *storepb.InfoRequest) (*storepb.InfoResponse, error) {
 	lset := p.externalLabels()
 	mint, maxt := p.timestamps()
 
 	res := &storepb.InfoResponse{
-		Labels:    make([]storepb.Label, 0, len(lset)),
+		Labels:    make([]labelpb.ZLabel, 0, len(lset)),
 		StoreType: p.component.ToProto(),
 		MinTime:   mint,
 		MaxTime:   maxt,
 	}
-	for _, l := range lset {
-		res.Labels = append(res.Labels, storepb.Label{
-			Name:  l.Name,
-			Value: l.Value,
-		})
-	}
+	res.Labels = append(res.Labels, labelpb.ZLabelsFromPromLabels(lset)...)
 
 	// Until we deprecate the single labels in the reply, we just duplicate
 	// them here for migration/compatibility purposes.
-	res.LabelSets = []storepb.LabelSet{}
+	res.LabelSets = []labelpb.ZLabelSet{}
 	if len(res.Labels) > 0 {
-		res.LabelSets = append(res.LabelSets, storepb.LabelSet{
+		res.LabelSets = append(res.LabelSets, labelpb.ZLabelSet{
 			Labels: res.Labels,
 		})
 	}
@@ -169,16 +155,16 @@ func (p *PrometheusStore) Series(r *storepb.SeriesRequest, s storepb.Store_Serie
 	}
 
 	if r.SkipChunks {
-		labelMaps, err := p.seriesLabels(s.Context(), newMatchers, r.MinTime, r.MaxTime)
+		labelMaps, err := p.client.SeriesInGRPC(s.Context(), p.base, newMatchers, r.MinTime, r.MaxTime)
 		if err != nil {
 			return err
 		}
 		for _, lbm := range labelMaps {
-			lset := make([]storepb.Label, 0, len(lbm)+len(externalLabels))
+			lset := make([]labelpb.ZLabel, 0, len(lbm)+len(externalLabels))
 			for k, v := range lbm {
-				lset = append(lset, storepb.Label{Name: k, Value: v})
+				lset = append(lset, labelpb.ZLabel{Name: k, Value: v})
 			}
-			lset = append(lset, storepb.PromLabelsToLabelsUnsafe(externalLabels)...)
+			lset = append(lset, labelpb.ZLabelsFromPromLabels(externalLabels)...)
 			sort.Slice(lset, func(i, j int) bool {
 				return lset[i].Name < lset[j].Name
 			})
@@ -211,7 +197,7 @@ func (p *PrometheusStore) Series(r *storepb.SeriesRequest, s storepb.Store_Serie
 
 	queryPrometheusSpan, ctx := tracing.StartSpan(s.Context(), "query_prometheus")
 
-	httpResp, err := p.startPromSeries(ctx, q)
+	httpResp, err := p.startPromRemoteRead(ctx, q)
 	if err != nil {
 		queryPrometheusSpan.Finish()
 		return errors.Wrap(err, "query Prometheus")
@@ -246,8 +232,7 @@ func (p *PrometheusStore) handleSampledPrometheusResponse(s storepb.Store_Series
 	span.SetTag("series_count", len(resp.Results[0].Timeseries))
 
 	for _, e := range resp.Results[0].Timeseries {
-		lset := p.translateAndExtendLabels(e.Labels, externalLabels)
-
+		lset := labelpb.ExtendLabels(labelpb.ZLabelsToPromLabels(e.Labels), externalLabels)
 		if len(e.Samples) == 0 {
 			// As found in https://github.com/thanos-io/thanos/issues/381
 			// Prometheus can give us completely empty time series. Ignore these with log until we figure out that
@@ -261,16 +246,13 @@ func (p *PrometheusStore) handleSampledPrometheusResponse(s storepb.Store_Series
 			continue
 		}
 
-		// XOR encoding supports a max size of 2^16 - 1 samples, so we need
-		// to chunk all samples into groups of no more than 2^16 - 1
-		// See: https://github.com/thanos-io/thanos/pull/718.
-		aggregatedChunks, err := p.chunkSamples(e, math.MaxUint16)
+		aggregatedChunks, err := p.chunkSamples(e, MaxSamplesPerChunk)
 		if err != nil {
 			return err
 		}
 
 		if err := s.Send(storepb.NewSeriesResponse(&storepb.Series{
-			Labels: lset,
+			Labels: labelpb.ZLabelsFromPromLabels(lset),
 			Chunks: aggregatedChunks,
 		})); err != nil {
 			return err
@@ -284,20 +266,16 @@ func (p *PrometheusStore) handleStreamedPrometheusResponse(s storepb.Store_Serie
 	level.Debug(p.logger).Log("msg", "started handling ReadRequest_STREAMED_XOR_CHUNKS streamed read response.")
 
 	framesNum := 0
-	seriesNum := 0
 
 	defer func() {
+		p.framesRead.Observe(float64(framesNum))
 		querySpan.SetTag("frames", framesNum)
-		querySpan.SetTag("series", seriesNum)
 		querySpan.Finish()
 	}()
 	defer runutil.CloseWithLogOnErr(p.logger, httpResp.Body, "prom series request body")
 
 	var (
-		lastSeries string
-		currSeries string
-		tmp        []string
-		data       = p.getBuffer()
+		data = p.getBuffer()
 	)
 	defer p.putBuffer(data)
 
@@ -319,19 +297,6 @@ func (p *PrometheusStore) handleStreamedPrometheusResponse(s storepb.Store_Serie
 
 		framesNum++
 		for _, series := range res.ChunkedSeries {
-			{
-				// Calculate hash of series for counting.
-				tmp = tmp[:0]
-				for _, l := range series.Labels {
-					tmp = append(tmp, l.String())
-				}
-				currSeries = strings.Join(tmp, ";")
-				if currSeries != lastSeries {
-					seriesNum++
-					lastSeries = currSeries
-				}
-			}
-
 			thanosChks := make([]storepb.AggrChunk, len(series.Chunks))
 			for i, chk := range series.Chunks {
 				thanosChks[i] = storepb.AggrChunk{
@@ -350,14 +315,16 @@ func (p *PrometheusStore) handleStreamedPrometheusResponse(s storepb.Store_Serie
 			}
 
 			if err := s.Send(storepb.NewSeriesResponse(&storepb.Series{
-				Labels: p.translateAndExtendLabels(series.Labels, externalLabels),
+				Labels: labelpb.ZLabelsFromPromLabels(
+					labelpb.ExtendLabels(labelpb.ZLabelsToPromLabels(series.Labels), externalLabels),
+				),
 				Chunks: thanosChks,
 			})); err != nil {
 				return err
 			}
 		}
 	}
-	level.Debug(p.logger).Log("msg", "handled ReadRequest_STREAMED_XOR_CHUNKS request.", "frames", framesNum, "series", seriesNum)
+	level.Debug(p.logger).Log("msg", "handled ReadRequest_STREAMED_XOR_CHUNKS request.", "frames", framesNum)
 	return nil
 }
 
@@ -421,7 +388,7 @@ func (p *PrometheusStore) chunkSamples(series *prompb.TimeSeries, maxSamplesPerC
 	return chks, nil
 }
 
-func (p *PrometheusStore) startPromSeries(ctx context.Context, q *prompb.Query) (presp *http.Response, err error) {
+func (p *PrometheusStore) startPromRemoteRead(ctx context.Context, q *prompb.Query) (presp *http.Response, _ error) {
 	reqb, err := proto.Marshal(&prompb.ReadRequest{
 		Queries:               []*prompb.Query{q},
 		AcceptedResponseTypes: p.remoteReadAcceptableResponses,
@@ -444,7 +411,9 @@ func (p *PrometheusStore) startPromSeries(ctx context.Context, q *prompb.Query)
 	}
 	preq.Header.Add("Content-Encoding", "snappy")
 	preq.Header.Set("Content-Type", "application/x-stream-protobuf")
-	preq.Header.Set("User-Agent", userAgent)
+	preq.Header.Set("X-Prometheus-Remote-Read-Version", "0.1.0")
+
+	preq.Header.Set("User-Agent", thanoshttp.ThanosUserAgent)
 	tracing.DoInSpan(ctx, "query_prometheus_request", func(ctx context.Context) {
 		preq = preq.WithContext(ctx)
 		presp, err = p.client.Do(preq)
@@ -472,18 +441,18 @@ func matchesExternalLabels(ms []storepb.LabelMatcher, externalLabels labels.Labe
 		return true, ms, nil
 	}
 
-	var newMatcher []storepb.LabelMatcher
-	for _, m := range ms {
-		// Validate all matchers.
-		tm, err := translateMatcher(m)
-		if err != nil {
-			return false, nil, err
-		}
+	tms, err := storepb.TranslateFromPromMatchers(ms...)
+	if err != nil {
+		return false, nil, err
+	}
 
-		extValue := externalLabels.Get(m.Name)
+	var newMatcher []storepb.LabelMatcher
+	for i, tm := range tms {
+		// Validate all matchers.
+		extValue := externalLabels.Get(tm.Name)
 		if extValue == "" {
 			// Agnostic to external labels.
-			newMatcher = append(newMatcher, m)
+			newMatcher = append(newMatcher, ms[i])
 			continue
 		}
 
@@ -498,7 +467,8 @@ func matchesExternalLabels(ms []storepb.LabelMatcher, externalLabels labels.Labe
 }
 
 // encodeChunk translates the sample pairs into a chunk.
-func (p *PrometheusStore) encodeChunk(ss []prompb.Sample) (storepb.Chunk_Encoding, []byte, error) {
+// TODO(kakkoyun): Linter - result 0 (github.com/thanos-io/thanos/pkg/store/storepb.Chunk_Encoding) is always 0.
+func (p *PrometheusStore) encodeChunk(ss []prompb.Sample) (storepb.Chunk_Encoding, []byte, error) { //nolint:unparam
 	c := chunkenc.NewXORChunk()
 
 	a, err := c.Appender()
@@ -506,98 +476,18 @@ func (p *PrometheusStore) encodeChunk(ss []prompb.Sample) (storepb.Chunk_Encodin
 		return 0, nil, err
 	}
 	for _, s := range ss {
-		a.Append(int64(s.Timestamp), float64(s.Value))
+		a.Append(s.Timestamp, s.Value)
 	}
 	return storepb.Chunk_XOR, c.Bytes(), nil
 }
 
-// translateAndExtendLabels transforms a metrics into a protobuf label set. It additionally
-// attaches the given labels to it, overwriting existing ones on collision.
-// Both input labels are expected to be sorted.
-//
-// NOTE(bwplotka): Don't use modify passed slices as we reuse underlying memory.
-func (p *PrometheusStore) translateAndExtendLabels(m []prompb.Label, extend labels.Labels) []storepb.Label {
-	pbLabels := storepb.PrompbLabelsToLabelsUnsafe(m)
-	pbExtend := storepb.PromLabelsToLabelsUnsafe(extend)
-
-	lset := make([]storepb.Label, 0, len(m)+len(extend))
-	ei := 0
-
-Outer:
-	for _, l := range pbLabels {
-		for ei < len(pbExtend) {
-			if l.Name < pbExtend[ei].Name {
-				break
-			}
-			lset = append(lset, pbExtend[ei])
-			ei++
-			if l.Name == pbExtend[ei-1].Name {
-				continue Outer
-			}
-		}
-		lset = append(lset, l)
-	}
-	for ei < len(pbExtend) {
-		lset = append(lset, pbExtend[ei])
-		ei++
-	}
-	return lset
-}
-
 // LabelNames returns all known label names.
-func (p *PrometheusStore) LabelNames(ctx context.Context, _ *storepb.LabelNamesRequest) (
-	*storepb.LabelNamesResponse, error,
-) {
-	u := *p.base
-	u.Path = path.Join(u.Path, "/api/v1/labels")
-
-	req, err := http.NewRequest("GET", u.String(), nil)
+func (p *PrometheusStore) LabelNames(ctx context.Context, r *storepb.LabelNamesRequest) (*storepb.LabelNamesResponse, error) {
+	lbls, err := p.client.LabelNamesInGRPC(ctx, p.base, r.Start, r.End)
 	if err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
+		return nil, err
 	}
-	req.Header.Set("User-Agent", userAgent)
-
-	span, ctx := tracing.StartSpan(ctx, "/prom_label_names HTTP[client]")
-	defer span.Finish()
-
-	resp, err := p.client.Do(req.WithContext(ctx))
-	if err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
-	}
-	defer runutil.ExhaustCloseWithLogOnErr(p.logger, resp.Body, "label names request body")
-
-	if resp.StatusCode/100 != 2 {
-		return nil, status.Errorf(codes.Internal, "request Prometheus server failed, code %s", resp.Status)
-	}
-
-	if resp.StatusCode == http.StatusNoContent {
-		return &storepb.LabelNamesResponse{Names: []string{}}, nil
-	}
-
-	var m struct {
-		Data   []string `json:"data"`
-		Status string   `json:"status"`
-		Error  string   `json:"error"`
-	}
-
-	body, err := ioutil.ReadAll(resp.Body)
-	if err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
-	}
-
-	if err = json.Unmarshal(body, &m); err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
-	}
-
-	if m.Status != SUCCESS {
-		code, exists := statusToCode[resp.StatusCode]
-		if !exists {
-			return nil, status.Error(codes.Internal, m.Error)
-		}
-		return nil, status.Error(code, m.Error)
-	}
-
-	return &storepb.LabelNamesResponse{Names: m.Data}, nil
+	return &storepb.LabelNamesResponse{Names: lbls}, nil
 }
 
 // LabelValues returns all known label values for a given label name.
@@ -609,124 +499,10 @@ func (p *PrometheusStore) LabelValues(ctx context.Context, r *storepb.LabelValue
 		return &storepb.LabelValuesResponse{Values: []string{l}}, nil
 	}
 
-	u := *p.base
-	u.Path = path.Join(u.Path, "/api/v1/label/", r.Label, "/values")
-
-	req, err := http.NewRequest("GET", u.String(), nil)
+	vals, err := p.client.LabelValuesInGRPC(ctx, p.base, r.Label, r.Start, r.End)
 	if err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
+		return nil, err
 	}
-	req.Header.Set("User-Agent", userAgent)
-
-	span, ctx := tracing.StartSpan(ctx, "/prom_label_values HTTP[client]")
-	defer span.Finish()
-
-	resp, err := p.client.Do(req.WithContext(ctx))
-	if err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
-	}
-	defer runutil.ExhaustCloseWithLogOnErr(p.logger, resp.Body, "label values request body")
-
-	if resp.StatusCode/100 != 2 {
-		return nil, status.Errorf(codes.Internal, "request Prometheus server failed, code %s", resp.Status)
-	}
-
-	if resp.StatusCode == http.StatusNoContent {
-		return &storepb.LabelValuesResponse{Values: []string{}}, nil
-	}
-
-	var m struct {
-		Data   []string `json:"data"`
-		Status string   `json:"status"`
-		Error  string   `json:"error"`
-	}
-	body, err := ioutil.ReadAll(resp.Body)
-	if err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
-	}
-
-	if err = json.Unmarshal(body, &m); err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
-	}
-
-	sort.Strings(m.Data)
-
-	if m.Status != SUCCESS {
-		code, exists := statusToCode[resp.StatusCode]
-		if !exists {
-			return nil, status.Error(codes.Internal, m.Error)
-		}
-		return nil, status.Error(code, m.Error)
-	}
-
-	return &storepb.LabelValuesResponse{Values: m.Data}, nil
-}
-
-// seriesLabels returns the labels from Prometheus series API.
-func (p *PrometheusStore) seriesLabels(ctx context.Context, matchers []storepb.LabelMatcher, startTime, endTime int64) ([]map[string]string, error) {
-	u := *p.base
-	u.Path = path.Join(u.Path, "/api/v1/series")
-	q := u.Query()
-
-	metric, err := matchersToString(matchers)
-	if err != nil {
-		return nil, errors.Wrap(err, "invalid matchers")
-	}
-
-	q.Add("match[]", metric)
-	q.Add("start", formatTime(timestamp.Time(startTime)))
-	q.Add("end", formatTime(timestamp.Time(endTime)))
-	u.RawQuery = q.Encode()
-
-	req, err := http.NewRequest("GET", u.String(), nil)
-	if err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
-	}
-	req.Header.Set("User-Agent", userAgent)
-
-	span, ctx := tracing.StartSpan(ctx, "/prom_series HTTP[client]")
-	defer span.Finish()
-
-	resp, err := p.client.Do(req.WithContext(ctx))
-	if err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
-	}
-	defer runutil.ExhaustCloseWithLogOnErr(p.logger, resp.Body, "series request body")
-
-	if resp.StatusCode/100 != 2 {
-		return nil, status.Errorf(codes.Internal, "request Prometheus server failed, code %s", resp.Status)
-	}
-
-	if resp.StatusCode == http.StatusNoContent {
-		return nil, nil
-	}
-
-	var m struct {
-		Data   []map[string]string `json:"data"`
-		Status string              `json:"status"`
-		Error  string              `json:"error"`
-	}
-
-	body, err := ioutil.ReadAll(resp.Body)
-	if err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
-	}
-
-	if err = json.Unmarshal(body, &m); err != nil {
-		return nil, status.Error(codes.Internal, err.Error())
-	}
-
-	if m.Status != SUCCESS {
-		code, exists := statusToCode[resp.StatusCode]
-		if !exists {
-			return nil, status.Error(codes.Internal, m.Error)
-		}
-		return nil, status.Error(code, m.Error)
-	}
-
-	return m.Data, nil
-}
-
-func formatTime(t time.Time) string {
-	return strconv.FormatFloat(float64(t.Unix())+float64(t.Nanosecond())/1e9, 'f', -1, 64)
+	sort.Strings(vals)
+	return &storepb.LabelValuesResponse{Values: vals}, nil
 }
